(*
 * Constraints.m
 *
 * Andrew D. Lewis
 * California Institute of Technology
 * 1 April 1996
 *
 * Mathematica for mechanical systems with constraints.
 *
 *)

BeginPackage["Constraints`",{"DerivativeOptions`","Affine`"}]

ConstrainedConnection::usage = "ConstrainedConnection[Conn,A,P,x,Deriv]
computes the constrained connection derived from the connection Conn and the
complementary projection P in coordinates x.  A is an arbitrary invertible
matrix function.";

Deform::usage = "Deform[X,Y,Conn,AConn,x] computes the image of X and Y
under the extra term in the constrained connection AConn.";

GeneralizedChristoffelSymbols::usage =
"GeneralizedChristoffelSymbols[X,g,conn,x] computes the generalized
Christoffel symbols for the basis of vector fields contained in the columns
of X.  Here g is the matrix for the Riemannian metric, conn are the
Christoffel symbols of the Levi-Civita connection, and x are the
coordinates."

MomentumEquation::usage = "MomentumEquation[g,X,gamma,x,xdot,Deriv] computes
the momentum equation for the vector field which is a linear combination of
the columns of the matrix X.  The Riemannian metric is g.";

MomentumMap::usage = "MomentumMap[g,X,gamma,x,xdot] computes the nonholonomic
momentum map with basis vector fields X with coefficients gamma.";

Begin["`Private`"]

Print["\nPackage \"Constraints\" defines: ConstrainedConnection, Deform, GeneralizedChristoffelSymbols, MomentumEquation, MomentumMap."];

Options[ConstrainedConnection] = {Deriv->Explicit}
ConstrainedConnection[Conn_, A_, P_, x_,opts___Rule] :=
                      Module[{opt1,local,DerivOpts},
(*
 * Gets the options for ConstrainedConnection.
 *)
  opt1 = Deriv /. {opts} /. Options[ConstrainedConnection];
  DerivOpts = {Explicit, Implicit};
  If[!MemberQ[DerivOpts, opt1],
    local = Indeterminate;
    Print["ConstrainedConnection: unknown option: ", opt1]
  ];
  If[opt1 == Explicit, local = ConstConn1[Conn,A,P,x]];
  If[opt1 == Implicit, local = ConstConn2[Conn,A,P,x]];
  local
  ];

ConstConn1[Conn_, A_, P_, x_] := Module[{B,Ainv},
(*
 * This module returns the Christoffel symbols of the constrained connection
 * associated with the connection Conn and the projection P onto a complement
 * of the constraint distribution.  Differentiation is explicit.
 *)
  B = A.P;
  Ainv = Inverse[A];
  Table[ Conn[[i]][[k]][[j]] + Sum[Ainv[[i]][[m]] D[B[[m]][[j]],x[[k]]],
             {m, Length[x]}] +
         Sum[Ainv[[i]][[m]] Sum[ Conn[[m]][[k]][[l]] B[[l]][[j]],
             {l, Length[x]}], {m, Length[x]}] -
         Sum[Ainv[[i]][[m]] Sum[ Conn[[m]][[k]][[j]] B[[i]][[l]],
             {l, Length[x]}], {m, Length[x]}],
         {i, Length[x]}, {j, Length[x]}, {k, Length[x]}]
  ];

ConstConn2[Conn_, A_, P_, x_] := Module[{B,Ainv},
(*
 * This module returns the Christoffel symbols of the constrained connection
 * associated with the connection Conn and the projection P onto a complement
 * of the constraint distribution.  Differentiation is implicit.
 *)
  B = A.P;
  Ainv = Inverse[A];
  Table[ Conn[[i]][[k]][[j]] + Sum[Ainv[[i]][[m]] Dt[B[[m]][[j]],x[[k]]],
             {m, Length[x]}] +
         Sum[Ainv[[i]][[m]] Sum[ Conn[[m]][[k]][[l]] B[[l]][[j]],
             {l, Length[x]}], {m, Length[x]}] -
         Sum[Ainv[[i]][[m]] Sum[ Conn[[m]][[k]][[j]] B[[i]][[l]],
             {l, Length[x]}], {m, Length[x]}],
         {i, Length[x]}, {j, Length[x]}, {k, Length[x]}]
  ];

GeneralizedChristoffelSymbols[X_,g_,conn_,x_] := Module[{n},
  n = Length[x];
  Table[(CovariantDerivative[X[[j]],X[[k]],conn,x].g.X[[i]])/
        (X[[i]].g.X[[i]]),{i,n},{j,n},{k,n}]
];

Deform[X_, Y_, Conn_, AConn_, x_] := Module[{},
(*
 * This module gives the value of the extra term in the constrained connection
 * on X and Y.
 *)
  Table[ Sum[ X[[j]] * Sum [AConn[[i]][[j]][[k]] Y[[k]],{k,Length[x]}],
           {j,Length[x]}] -
         Sum[ X[[j]] * Sum [Conn[[i]][[j]][[k]] Y[[k]],{k,Length[x]}],
           {j,Length[x]}], {i,Length[x]}]
  ];

Options[MomentumEquation] = {Deriv->Explicit}
MomentumEquation[g_, X_, gamma_, x_, xdot_, opts___Rule] :=
                 Module[{opt1,local,DerivOpts},
(*
 * Gets the options for MomentumEquation.
 *)
  opt1 = Deriv /. {opts} /. Options[MomentumEquation];
  DerivOpts = {Explicit, Implicit};
  If[!MemberQ[DerivOpts, opt1],
    local = Indeterminate;
    Print["MomentumEquation: unknown option: ", opt1]
  ];
  If[opt1 == Explicit, local = MomentumEq1[g,X,gamma,x,xdot]];
  If[opt1 == Implicit, local = MomentumEq2[g,X,gamma,x,xdot]];
  local
  ];

MomentumEq1[g_, X_, gamma_, x_, xdot_] := Module[{temp1,temp2},
(*
 * Computes the momentum equation.  Differentiation is explicit.
 *)
  temp1 = Table[ Sum[ D[gamma[[a]],x[[k]]] xdot[[k]], {k, Length[x]}],
                {a, Length[gamma]}];
  temp2 = Table[ Sum[ temp1[[a]] X[[a]][[i]], {a, Length[gamma]}],
                {i, Length[x]}];
  temp2.g.xdot
  ];

MomentumEq2[g_, X_, gamma_, x_, xdot_] := Module[{temp1,temp2},
(*
 * Computes the momentum equation.  Differentiation is implicit.
 *)
  temp1 = Table[ Sum[ Dt[gamma[[a]],x[[k]]] xdot[[k]], {k, Length[x]}],
                {a, Length[gamma]}];
  temp2 = Table[ Sum[ temp1[[a]] X[[a]][[i]], {a, Length[gamma]}],
                {i, Length[x]}];
  temp2.g.xdot
  ];

MomentumMap[g_, X_, gamma_, x_, xdot_] := Module[{temp1,a,alpha},
(*
 * Computes the nonholonomic momentum map.
 *)
  temp1 = Table[Sum[X[[alpha]][[j]] gamma[[alpha]][[a]],
                {alpha, Length[gamma]}],
                {j, Length[x]}, {a, Length[Transpose[gamma]]}];
  Table[Sum[Sum[g[[i]][[j]] xdot[[i]] temp1[[j]][[a]],{j, Length[x]}],
        {i, Length[x]}], {a, Length[Transpose[gamma]]}]
  ];

End[]
EndPackage[]
